// ===============================================
// File:                Objects.cxx
// Description:         VRML-Nodes defined with OpenGL
// Current:             Sept 28th 1999
// Author(s):           Thomas Kern
// ===============================================

/*
 *        thk VRML-Viewer (based on OpenGL)
 *  it shows static *.wrl files based on the VRML97 standard
 *
 *          Copyright (C) 1997-1999  Thomas Kern
 *                  All Rights Reserved
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * My email adress is: 
 *   Thomas.Kern@3Dsource.de 
 * 
 * You should find these files at: my homepage
 *   http://www.3Dsource.de
 *
 */

//=====================================================================
//	Description
//=====================================================================
// 
/*
  These objectes are used by VRMLControl
  Some funftions are located in Tools.h/.cxx.
*/


//=====================================================================
//      INCLUDES
//=====================================================================
//
#ifndef OBJECTS_H
#define OBJECTS_H
//---------------------------------------------------------------------
#include <stdio.h>

#include "Tools.H"            // fuer X,Y,Z

#define  MAX_CHILDREN      30   // Anzahl der Children innerhalb []
#define  MAX_ECKEN         20   // Kreis wird durch n=20+1 Polygone dargestellt
#define  USE_STATIC_FIELDS      // Coordinaten-Felder mit fester Groesse
#define  MAXVERTICES     4000   // Punkte(X,Y,Z) pro Polygonzug
#define  INDEX_MULT         5   // Anzahl der Indizes auf den gleichen Punkt

enum NodeType { Appearance, Box, Color, Cone, Coordinate, Cylinder,
                DirectionalLight, Group, ImageTexture, IndexedFaceSet,
                IndexedLineSet,
                Material, Normal, PixelTexture, PointLight, Shape,
                Sphere, TextureCoordinate, Transform,
                TextureTransform, notValidNode };

class AppearanceNode;           // forward declaration

//=====================================================================
// Basisklasse einer VRML-Node.
// Diese Klasse ermoeglicht das einheitliche Ansprechen der Nodes.
// Jede VRMLNode kann ueber DEF/USE benannt und unter diesem Namen
// mehrfach benutzt werden.
//=====================================================================
//
class VRMLNode
{
  public:
    VRMLNode() {
      myNodeType = notValidNode;
      drawWithTexture = true;
      NodeName[0]     = '\0';
    };
    virtual ~VRMLNode() {};

    void SetName(char myName[]) {                       // kann ueber einen bestimmten Namen
                 sprintf(NodeName, "%s", myName);       // angesprochen werden
    };
    void GetName(char myName[]) {
      sprintf(myName, "%s",NodeName);
    };
    NodeType GetNodeTyp() {
      return myNodeType;
    };
    void SetTexture(boolean On) {
      drawWithTexture = On;
    };

    virtual void DrawNode() = 0;

  protected:
    NodeType    myNodeType;
    boolean     drawWithTexture;

  private:
    char        NodeName[WORD_LENGHT+1];    // Name dieser VRML-Node
};

//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Grouping-Node.
// Diese Klasse enthaelt Verweise auf andere Nodes (Childrens).
//=====================================================================
//
class GroupingNode : public VRMLNode
{

  public:
    GroupingNode() {
      MyChildrenCount = 0;
    };
    virtual ~GroupingNode() {};

    void  AddChildren(VRMLNode *Children);

    virtual void DrawNode() = 0;

    int       MyChildrenCount;
    VRMLNode *MyChildren[MAX_CHILDREN];
};

//=====================================================================
// Von der allgemeinen VRML-Node abgeleitetes GeometryNode.
// Diese Klasse ist eine Basisklasse fuer alle geometrischen Objekte.
//=====================================================================
//
class GeometryNode : public VRMLNode
{
  public:
    GeometryNode() {
      haveAppearance = false;
    };
    virtual ~GeometryNode() {};

    void AddAppearance(AppearanceNode *Appearance) {
      MyAppearance   = Appearance;              // Material des Objekts
      haveAppearance = true;
    };

    virtual void DrawNode() = 0;

    AppearanceNode *MyAppearance;
    boolean         haveAppearance;
};

//=====================================================================
// Von der allgemeinen Grouping-Node abgeleitete Group-Node.
// Diese Klasse enthaelt keine eigene Funktionalitaet, sondern nur
// Verweise auf andere Nodes.
//=====================================================================
//
class GroupNode : public GroupingNode
{
  public:
    GroupNode() {
      myNodeType      = Group;
      MyChildrenCount = 0;
    };
    virtual ~GroupNode() {};

    virtual void DrawNode();
};

//=====================================================================
// Von der allgemeinen Grouping-Node abgeleitete Transform-Node.
// Diese Klasse erledigt die Transformationen (Rotation, Translation)
// und enthaelt Verweise auf die betroffenen Nodes.
//=====================================================================
//
class TransformNode : public GroupingNode
{
  public:
    TransformNode();
    virtual ~TransformNode() {};

    void  Center(GLfloat Axle[3]);
    void  Rotation(GLfloat Angle, GLfloat Axle[3]);  // in degrees
    void  Scale(GLfloat Axle[3]);
    void  ScaleOrientation(GLfloat Angle, GLfloat Axle[3]);
    void  Translation(GLfloat Axle[3]);

    virtual void DrawNode();

  private:
    GLfloat   CenM[3], RotM[4], ScaleM[3], ScaleOrM[4], TransM[3];
    boolean   center, rotation , scale, scaleOr, translation;
};

//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete TextureTransform-Node.
// Diese Klasse verzerrt das Texturbild eines Objektes.
//=====================================================================
//
class TextureTransformNode : public VRMLNode
{
  public:
    TextureTransformNode();
    virtual ~TextureTransformNode() {};

    void  Center(GLfloat Axle[2]);
    void  Rotation(GLfloat Angle);  // in degrees
    void  Scale(GLfloat Axle[2]);
    void  Translation(GLfloat Axle[2]);

    virtual void DrawNode();

  private:
    GLfloat   CenM[2], RotV, ScaleM[3], TransM[3];
    boolean   center, rotation , scale, translation;
};

//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Texture-Node.
// Diese Klasse beschreibt das Texturbild eines Objektes.
//=====================================================================
//
class TextureNode : public VRMLNode
{
  public:
    TextureNode();
    virtual ~TextureNode();

    void AddRepeatS(boolean RepeatS) {
      repeatS = RepeatS;
    };
    void AddRepeatT(boolean RepeatT) {
      repeatT = RepeatT;
    };
    void SetModulate(boolean set) {
      modulate = set;
    };
    int  GetTextureID() {
      return textureID;
    };
    void PrepareTexture();

    virtual void DrawNode();

    unsigned int    textureID;
    boolean         repeatS, repeatT, modulate;
    unsigned char  *ImagePtr;
    int             Breite, Hoehe;
    GLclampf          Color4[4];
    boolean         madeTexture, IsReady;
};

//=====================================================================
// Von der Texture-Node abgeleitete ImageTexture-Node.
//=====================================================================
//
class ImageTextureNode : public TextureNode
{
  public:
    ImageTextureNode(int texturNr);
    virtual ~ImageTextureNode() {};

    void AddTextureURL(char FileName[]) {
      sprintf(url, "%s", FileName);
      ScanTextureImage();
    };

  private:
    void ScanTextureImage();

    char url[URL_SIZE];
};

//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Material-Node.
// Diese Klasse beschreibt die Materialeigenschaften von Objekten.
//=====================================================================
//
class MaterialNode : public VRMLNode
{
  public:
    MaterialNode();
    virtual ~MaterialNode() {};

    void UseMaterial(boolean lightIsOn) {
      useMaterial = lightIsOn;
    };
    void AddAmbient(     GLclampf Intensity);
    void AddDiffuse(     GLclampf Color[3] );
  //    void AddEmissive(    GLclampf Color[3] );
  //    void AddShininess(   GLclampf Intensity);
  //    void AddSpecular(    GLclampf Color[3] );
    void AddTransparency(GLclampf Intensity);

    virtual void DrawNode();

  private:
    boolean   useMaterial;
    GLclampf  AmbientMult;
    GLclampf  AmbientColor4[4];
    GLclampf  DiffuseColor4[4];
};

//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Appearance-Node.
// Diese Klasse beschreibt das Aussehen von Objekten.
//=====================================================================
//
class AppearanceNode : public VRMLNode
{
  public:
    AppearanceNode() {
      haveTexture = haveMaterial = false;
      haveTextureTransform       = false;
      ownMaterial = ownTextureTransform = false;
      myNodeType  = Appearance;
    };
    virtual ~AppearanceNode() {
      if( ownMaterial )
        delete MyMaterial;
      if( ownTextureTransform )
        delete MyTextureTransform;
    };

    void AddMaterial(MaterialNode *Material) {
      MyMaterial   = Material;
      haveMaterial = true;
    };

    void AddTexture(TextureNode *Texture) {
      MyTexture   = Texture;
      haveTexture = true;
    };

    void AddTextureTransform(TextureTransformNode *TextureTransform) {
      MyTextureTransform   = TextureTransform;
      haveTextureTransform = true;
    };


    virtual void DrawNode();

  private:
    TextureNode           *MyTexture;
    TextureTransformNode  *MyTextureTransform;
    MaterialNode          *MyMaterial;
    boolean                haveTexture, haveTextureTransform, haveMaterial;
    boolean                ownMaterial, ownTextureTransform;
};

//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Field-Node.
// Diese Klasse stellt ein Datenfeld float-Werten zur Verfuegung.
// Erzeugt wird das Datenfeld vom Aufrufer direkt aus einer Datei.
// Diese Klasse kann fuer die Nodes: Color, Coordinate, Normal
// und TextureCoordinate angewendet werden. Zur eindeutigen Zuordnung
// sollte der Typ explizit gesetzt werden (SetNodeTyp).
//=====================================================================
//
class FieldNode : public VRMLNode
{
  public:
    FieldNode();
    virtual ~FieldNode();

    void SetNodeTyp(NodeType ThisNodeTyp) {
      myNodeType = ThisNodeTyp;
    };

    void NewField(GLfloat *FieldPtr, int FieldSize);

    GLfloat *ReturnFieldPtr() {
      return myValueField;
    };

    int  ReturnVertexCount() {
      return VertexCount;
    };

    virtual void DrawNode() {};

  private:
#ifdef USE_STATIC_FIELDS
    GLfloat      myValueField[3*MAXVERTICES];
#else    
    GLfloat     *myValueField;
#endif
    int          VertexCount;
};

//=====================================================================
// Von einer GeometryNode abgeleitete Box-Node.
// Diese Klasse beschreibt einen Quader.
//=====================================================================
//
class BoxNode : public GeometryNode
{
  public:
    BoxNode();
    virtual ~BoxNode() {};

    void AddSize(GLfloat GeomMatrix[3]);
    virtual void DrawNode();

  private:
    void DrawWithTexture();

    GLfloat      Plu[3], Pru[3], Pro[3], Plo[3]; // Punkte Front
    GLfloat      Qlu[3], Qru[3], Qro[3], Qlo[3]; // Punkte Back
    GLfloat      Tlu[2], Tru[2], Tro[2], Tlo[2]; // Punkte Textur
    GLfloat      Nfl[6][3];                      // Normalenvektoren
};

//=====================================================================
// Von einer GeometryNode abgeleitete Cone-Node.
// Diese Klasse beschreibt einen Kegel.
//=====================================================================
//
class ConeNode : public GeometryNode
{
  public:
    ConeNode();
    virtual ~ConeNode() {};

    void AddRadius(GLfloat Radius) {
      R = Radius;
      CalcGeometry();
    };
    void AddHeight(GLfloat Height) {
      H = Height;
      CalcGeometry();
    };
    void AddSide(boolean Side) {
      drawSide = Side;
    };
    void AddBottom(boolean Bottom) {
      drawBottom = Bottom;
    };

    virtual void DrawNode();

  private:
    void  CalcGeometry();
    void  DrawWithTexture();

    GLfloat      R, H;
    boolean      drawSide, drawBottom;
    GLfloat      P[MAX_ECKEN][3], Q[3], U[3];  // Koordinaten
    GLfloat      Nfl[MAX_ECKEN+1][3];          // Normalenvektoren
    int          split;
};

//=====================================================================
// Von einer GeometryNode abgeleitete Cylinder-Node.
// Diese Klasse beschreibt einen Zylinder.
//=====================================================================
//
class CylinderNode : public GeometryNode
{
  public:
    CylinderNode();
    virtual ~CylinderNode() {};

    void AddRadius(GLfloat Radius) {
      R = Radius;
      CalcGeometry();
    };
    void AddHeight(GLfloat Height) {
      H = Height;
      CalcGeometry();
    };
    void AddSide(boolean Side) {
      drawSide = Side;
    };
    void AddBottom(boolean Bottom) {
      drawBottom = Bottom;
    };
    void AddTop(boolean Top) {
      drawTop = Top;
    };

    virtual void DrawNode();

  private:
    void  CalcGeometry();
    void  DrawWithTexture();

    GLfloat       R, H;
    boolean       drawSide, drawTop, drawBottom;
    GLfloat       P[MAX_ECKEN][3], Q[MAX_ECKEN][3];  // Koordinaten
    GLfloat       UP[3], UQ[3];
    GLfloat       Nfl[MAX_ECKEN+2][3];               // Normalenvektoren
    int           split;
};

//=====================================================================
// Von einer GeometryNode abgeleitete Sphere-Node.
// Diese Klasse beschreibt eine Kugel.
//=====================================================================
//
class SphereNode : public GeometryNode
{
  public:
    SphereNode();
    virtual ~SphereNode();

    void AddRadius(GLfloat Radius) {
      R = Radius;
      CalcGeometry();
    };

    virtual void DrawNode();

  private:
    void  CalcGeometry();
    void  DrawWithTexture();

    GLUquadricObj     *Quadric;
    GLfloat            R;
    int                split;
};

//=====================================================================
// Von einer GeometryNode abgeleitete FaceSet-Node.
// Diese Klasse beschreibt einen Polygonzug.
// Wichtig: Gemaess der Spezifikation wird vorausgesetzt, dass
// die Datenfelder der Groesse ihrem Indexfeld entsprechen und
// vollstaendig sind. Eine Ueberpruefung findet aus Performance-
// gruenden nicht mehr statt !
//=====================================================================
//
class FaceSetNode : public GeometryNode
{
  public:
    FaceSetNode();
    virtual ~FaceSetNode();

    void AddCoordField(GLfloat *CoordField);
    void AddCoordIndex(int *CoordIndex, int CoordIndexSize);
    void AddColorField(GLfloat *ColorField);
    void AddColorIndex(int *ColorIndex, int ColorIndexSize);
    void AddNormalField(GLfloat *NormalField);
    void AddTextureField(GLfloat *TextureCoordField);
    void AddTextureIndex(int *TextureCoordIndex, int TextureIndexSize);

    void SetCCW(boolean ccw) {
      CCW = ccw;
    };
    void SetConvex(boolean convex) {
      Convex = convex;
    };
    void SetColorPerVertex(boolean colorPerVertex) {
      ColorPerVertex = colorPerVertex;
    };
    void SetNormalPerVertex(boolean normalPerVertex) {
      NormalPerVertex = normalPerVertex;
    };
    void SetCreaseAngle(GLfloat Angle) {        // in radians
      creaseAngle = Angle;
    };

    void PrepareForDraw();

    virtual void DrawNode() {
      if(Convex)
        DrawConvexOnly();
      else                                    // nicht stabil
        DrawTesselated();
      glDisable(GL_TEXTURE_2D);
    };

  private:
    void CalcNormals();
    void DrawTesselated();
    void DrawConvexOnly();

    GLfloat     *MyCoords, *MyColors, *MyNormals, *MyTextureCoords;
    boolean      haveCoords, haveColors, haveNormals, haveTextureCoords;
    boolean      haveOwnNormals;
#ifdef  USE_STATIC_FIELDS      // Koordinaten-Felder mit fester Groesse
    int          MyCoordIndex[INDEX_MULT*MAXVERTICES];
    int          MyColorIndex[INDEX_MULT*MAXVERTICES];
    int          MyTextureIndex[INDEX_MULT*MAXVERTICES];
#else
    int         *MyCoordIndex, *MyColorIndex, *MyTextureIndex;
#endif
    int          MyCoordIndexSize, MyColorIndexSize, MyTextureIndexSize;
    boolean      haveCoordIndex, haveColorIndex, haveTextureIndex;
    boolean      CCW, ColorPerVertex, Convex, NormalPerVertex;
    GLfloat      creaseAngle;
    enum { nothingToDraw = 0,
           drawNormalPerVertex =  1, drawNormalPerFace =  2,
           drawColorPerVertex  = 10, drawColorPerFace   = 20,
           drawTexture         = 5  } myDrawing;
    int    whichDrawing;
};

//=====================================================================
// Von einer GeometryNode abgeleitete LineSet-Node.
// Diese Klasse beschreibt einen Polygonzug aus Linien.
// Wichtig: Gemaess der Spezifikation wird vorausgesetzt, dass
// die Datenfelder der Groesse ihrem Indexfeld entsprechen und
// vollstaendig sind. Eine Ueberpruefung findet aus Performance-
// gruenden nicht mehr statt !
//=====================================================================
//
class LineSetNode : public FaceSetNode
{
  public:
    LineSetNode();
    virtual ~LineSetNode();

    void AddCoordField(GLfloat *CoordField);
    void AddCoordIndex(int *CoordIndex, int CoordIndexSize);
    void AddColorField(GLfloat *ColorField);
    void AddColorIndex(int *ColorIndex, int ColorIndexSize);

    void SetColorPerVertex(boolean colorPerVertex) {
      ColorPerVertex = colorPerVertex;
    };

    void PrepareForDraw();

    virtual void DrawNode() {
      DrawConvexOnly();
      glDisable(GL_TEXTURE_2D);
    };

  private:
    //void CalcNormals();
    void DrawConvexOnly();

    GLfloat         *MyCoords, *MyColors;
    boolean      haveCoords, haveColors;
#ifdef  USE_STATIC_FIELDS      // Koordinaten-Felder mit fester Groesse
    int          MyCoordIndex[INDEX_MULT*MAXVERTICES];
    int          MyColorIndex[INDEX_MULT*MAXVERTICES];
#else
    int         *MyCoordIndex, *MyColorIndex;
#endif
    int          MyCoordIndexSize, MyColorIndexSize;
    boolean      haveCoordIndex, haveColorIndex;
    boolean      ColorPerVertex;
    enum { nothingToDraw = 0,
           drawColorPerVertex  = 10, drawColorPerFace   = 20,
         } myDrawing;
    int  whichDrawing;
};

//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Shape-Node.
// Diese Klasse enthaelt Verweise auf Geometrie und Material von Objekten.
//=====================================================================
//
class ShapeNode : public VRMLNode
{
  public:
    ShapeNode() {
      haveAppearance = haveGeometry = false;
      myNodeType     = Shape;
    };
    virtual ~ShapeNode() {};

    void  AddAppearance(AppearanceNode *Appearance) {
      MyAppearance   = Appearance;
      haveAppearance = true;
    };
    void  AddGeometry(GeometryNode *Geometry) {
      MyGeometry   = Geometry;
      haveGeometry = true;
    };

    void AskForAppearance() {
      if( haveAppearance && haveGeometry )
        MyGeometry->AddAppearance(MyAppearance);
    };

    virtual void DrawNode() {
      if(haveGeometry)
        MyGeometry->DrawNode();
    };

  private:
    AppearanceNode  *MyAppearance;
    GeometryNode    *MyGeometry;
    boolean          haveAppearance, haveGeometry;
};


//=====================================================================
// nicht durch den Standard definierte Ergaenzungen


//=====================================================================
//  Koordinatensystem
//=====================================================================
//
class GLCoordSystem
{
  public:
    GLCoordSystem();
    ~GLCoordSystem() {};

    void DrawGLObjectAsList();

  private:
    GLfloat             CoordColor[3], CoordV0[3], CoordV1[3];
    GLfloat             CoordA1[3], CoordA2[3];
    boolean             CoordList;
    int                 CoordListNr;
};

//==============================================================================
#endif // OBJECTS_H

