// ===============================================
// File:          VRMLControl.C/.H
// Description:   Parser for VRML97-files 
// Version 1:     
// Current:       
// Author(s):     Thomas Kern
// ===============================================

// Diese Klasse abstrahiert gegenueber dem Aufrufer eigenstaendige
// Objekte. Der Aufrufer uebergibt den Dateinamen einer VRML-Datei,
// und kann dann das Zeichnen aller darin enthaltenen Objekte
// (meist mehrere Koerper) ueber einen einzigen Zugriff erwarten.
// Damit das starre VRML-Objekt effektiv gezeichnet wird, lassen
// sich die "Unterobjekte" in einer Diplay-List zusammenfassen.
//

/*
 * Copyright (C) 1997-2003  Thomas Kern
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * My email adress is: 
 *   Thomas.Kern@3Dsource.de 
 * 
 * You should find these files at: my homepage
 *   http://www.3Dsource.de
 */

//=====================================================================
//	Description
//=====================================================================
// 
/*
	Here you can find the VRMLControl class.
  It is a wrapper class for the OpenGL objects in Objects.cxx.
  Some funftions are located in Tools.h/.cxx.

  You have to call the constructor with a *.wrl file name as a parameter.

  The constructor calls "ScanFile();", where
	- the file name and the directory names are separated (functions from Tools.h/.cxx)
	  + CheckFile(DateiName, "r")
	  + extractDir(char *complete, char *directory, char *file)
	- the directories are saved for file searching operations 
	  + scannedDir[dirCount]
	- are checks for the right file type (only unpacked VRML 2.0 are supported)
	  + CheckForType()
	- are the calls for parsing the file
	  + ScanForRootNodes();

  The method "ScanForRootNodes();"  
  - scans for "root nodes" (spezific VRML object entries) and generate node-names 
    + GenNodeName(NextNodeName);
  - scans for specific entries (not VRML, only my entries and comments)
    + #thk:NoTesselation, #thk:NoDecalTexture, #
  - scans for group or transformation entries:
    + if( !strcmp(nextString, "Group") )
        AddGroup(Ebene, Parent, NextNodeName);
    + else if( !strcmp(nextString, "Transform") )
        AddTransform(Ebene, Parent, NextNodeName);

  The method "AddGroup(Ebene, Parent, NextNodeName);" (objects from Objects.h/.cxx)
  - creates a GroupNode() object, name it and store the object in a list
    + ThisNode = new GroupNode();
    + ThisNode->SetName(ThisNodeName);
    + if(ObjectCopyCount < MAXNODES) {
        ObjectPtr[ObjectCount] = ThisNode;
  - specify the node as a root node (level 0) or a child (this node is encapsulated in another node entry)
    + RootNode[RootNodeCount] = ThisNode;
    + Parent->AddChildren( (VRMLNode*)ThisNode );
  - scans the node entry
    + ScanForBraces(Klammer2); // "{" increase a counter, "}" decrease the counter
    + adds a child node (keyword "children")
      AddChildren( Ebene, (GroupingNode*)ThisNode );
    + returns if "closing braces - open braces" = NULL

  The method "AddTransform(Ebene, Parent, NextNodeName);" (objects from Objects.h/.cxx)
  - creates a TransformNode() object
  - is looking for transform keywords ("center", "rotation", "translation", "scale" etc.)
  - calls private methods, i.e.
    + ThisNode->Rotation(Angle, Axle);

  The method "AddChildren(int Ebene, GroupingNode *Parent)"
  - counts the opening and closing brackets
    + ScanForBrackets(Klammer1); // []
  - is looking for keywords ("USE", "DEF", "Shape", "Group", Transform")
  - calls recursive methods
    + AddShape, AddGroup, AddTransform

  The method "AddShape(GroupingNode *Parent, char ThisNodeName[])" (objects from Objects.h/.cxx)
  - creates a ShapeNode() object
  - is looking for keywords and calls shape functions
    + KeyAppearance(ThisNode, NextNodeName);
    + KeyGeometry(ThisNode, NextNodeName);
  - or if there isn't an appearance entry, calls
    + AskForAppearance();
  
  The method "KeyAppearance(ShapeNode *MyShape, char ThisNodeName[])"
  - gets appearance entries from parent nodes ("USE")
  - scans for new appearance entries
    + AddAppearance(MyShape, ThisNodeName);

  The method "AddAppearance( ShapeNode *MyShape, char ThisNodeName[] )"
  - creates an AppearanceNode() object
  - adds an appearance entry to the parent shape node
    + MyShape->AddAppearance(ThisNode);
  - gets appearance entries from parent nodes ("USE")
    + ThisNode->AddMaterial, AddTexture, AddTextureTransform
  - scans the material, texture or texture transform entries
    + KeyMaterial, KeyTexture, KeyTextureTransform

  The method "AddMaterial(AppearanceNode *MyAppearance, char ThisNodeName[])"
  - creates a MaterialNode() object
  - adds a material entry to the parent appearance node
    + MyAppearance->AddMaterial(ThisNode);
  - scans and stores color and transparency values (OpenGL ready)

  The method "AddImageTexture(AppearanceNode *MyAppearance, char ThisNodeName[])"
  - creates an ImageTextureNode(TextureNr)
    + TextureNr = texture ID
    + AddTextureURL(url) (url = file name)
    + ScanTextureImage() (from Objects.h/.cxx)
    + loadPPMfile and loadJPEGfile (loading routines for PPM and JPEG from Tools.h/.cxx)
    + PrepareTexture() (OpenGL setup for this texture ID)

  The method "AddTextureTransform( AppearanceNode *MyAppearance, char ThisNodeName[] )"
  - creates a TextureTransformNode() for rotation, translation an scaling of a texture

  The method "KeyGeometry( ShapeNode *MyShape, char ThisNodeName[] )"
  - scans for different geometry keywords like
    + Cone, Cylinder, Sphere, Box (simple geometry objects)
    + IndexedFaceSet and IndexedLineSet (complex geometry objects)

  The Add-methods for Cone, Cylinder, Sphere, Box
  - create geometry node objects (from Objects.h/.cxx)
  - add a geometry entry to the parent appearance node
    + MyShape->AddGeometry( (GeometryNode*)ThisNode );

  The method AddLineSet
  - creates a Line Set object and adds a geometry entry to the parent appearance node
  - loads arrays for color, coordinates and the indices
    + KeyField(ThisNode, Color, NextNodeName);
    + KeyField(ThisNode, Coordinate, NextNodeName);
    + AddIndex(ThisNode, Color);
    + AddIndex(ThisNode, Coordinate);
  - calls PrepareForDraw() for setup 

  The method AddFaceSet
  - creating and array loading is like a Line Set Node
  - loads some extra settings (like TesselationFlag)
  - calls PrepareForDraw() for setup and calculating the normal vectors

*/

//=====================================================================
//      INCLUDES
//=====================================================================
//
#ifndef VRMLCONTROL_H
#define VRMLCONTROL_H
//---------------------------------------------------------------------
#include <stdio.h>

#include "Objects.H"
#include "Tools.H"

//=====================================================================
//      eigene Define's
//=====================================================================
//
#define  MAXNODES               3000  // begrenzte Anzahl an Nodes

class AppearanceNode;                 // forward declarations
class GroupingNode;
class ShapeNode;
class VRMLNode;

//=====================================================================
// Klassendeklaration
//=====================================================================
//
// Damit die Verschachtelung der einzelnen Methoden nachvollzogen
// werden kann, ist eine Einarbeitung in den ISO-Standard der
// VRML97 zu empfehlen.
//
class VRMLControl
{
  public:
    VRMLControl(char DateiName[]);
    ~VRMLControl();

    void DrawObject();                          // zu Testzwecken
    void DrawAsList();                          // sonst DrawingList verwenden
    void SetTexture(boolean On);                // Texturen nutzen

  private:
    void  GenNodeName(char newName[]);          // einen Namen erzeugen
    int   UseNodeName();                        // Objekt wiederverwenden
    void  ScanForBraces(int &Klammer2);         // geschweifte Klammern suchen
    void  ScanForBrackets(int &Klammer1);       // eckige Klammern suchen
    void  ScanURL(char DateiName[]);            // URL fuer Textur
    void  AddBox( ShapeNode *MyShape,
                  char ThisNodeName[] );        // Wuerfel
    void  AddCone( ShapeNode *MyShape,
                   char ThisNodeName[] );       // Kegel
    void  AddCylinder( ShapeNode *MyShape,
                       char ThisNodeName[]);    // Zylinder
    void  AddSphere( ShapeNode *MyShape,
                     char ThisNodeName[]);      // Kugel
    void  AddIndex( FaceSetNode *MyFaceSet,
                    NodeType MyNodeType );      // Feld mit Indizes
    void  AddField( FaceSetNode *MyFaceSet,
                    NodeType     MyNodeType,
                    char         ThisNodeName[] ); // Datenfeld
    void  KeyField( FaceSetNode *MyFaceSet,
                    NodeType     MyNodeType,
                    char         ThisNodeName[] ); // Verteiler
    void  AddFaceSet( ShapeNode *MyShape,
                      char ThisNodeName[] );       // Polygonzug
    void  AddLineSet( ShapeNode *MyShape,
                      char ThisNodeName[] );       // Linienzug

    void  KeyGeometry( ShapeNode *MyShape,
                       char ThisNodeName[] );      // Verteiler
    void  AddTextureTransform( AppearanceNode *MyAppearance,
                               char ThisNodeName[] ); // Textur transformieren
    void  KeyTextureTransform( AppearanceNode *MyAppearance,
                               char ThisNodeName[] ); // Verteiler
    void  AddImageTexture( AppearanceNode *MyAppearance,
                           char ThisNodeName[] );     // Textur aus Datei
    void  KeyTexture( AppearanceNode *MyAppearance,
                      char ThisNodeName[] );          // Verteiler
    void  AddMaterial( AppearanceNode *MyAppearance,
                       char ThisNodeName[] );         // Material
    void  KeyMaterial( AppearanceNode *MyAppearance,
                       char ThisNodeName[] );         // Verteiler
    void  AddAppearance( ShapeNode *MyShape,
                         char ThisNodeName[] );       // gesamtes Aeusseres
    void  KeyAppearance( ShapeNode *MyShape,
                         char ThisNodeName[] );       // Verteiler
    void  AddShape( GroupingNode *Parent,
                    char ThisNodeName[] );            // vollst. Koerper
    void  AddChildren( int Ebene,
                       GroupingNode *Parent );        // Verteiler
    void  AddTransform( int Ebene, GroupingNode *Parent,
                        char ThisNodeName[] );        // Transformation
    void  AddGroup( int Ebene, GroupingNode *Parent,
                    char ThisNodeName[] );            // Sammelstelle
    void  ScanForRootNodes();
    int   CheckForType();                             // VRML1, 2 oder Inventor
    void  ScanFile();                                 // Datei auswerten

    // beschreibende wrl-Datei
    FILE          *file;                           // beschreibende Datei
    char           DateiName[128];

    // abstrakte Klasse beschreibt ein  VRML-Objekt
    VRMLNode     *ObjectPtr[MAXNODES];             // dynamisch angelegtes Feld
    VRMLNode     *ObjectCopyPtr[MAXNODES];         // mehrfache Verwendung
    VRMLNode     *RootNode[50];                    // nur diese lassen sich zeichnen
    int           ObjectCount;                     // fuer delete
    int           ObjectCopyCount;                 // Verwaltung
    int           RootNodeCount;                   // fuer's Zeichnen

    // Vertex-Arrays
    GLfloat       *VertexFieldPtr[MAXNODES];       // Zeiger auf neues Feld
    int            VFieldCount;                    // Anzahl der Felder

    // Texturverwaltung
    unsigned       TextureNr;                      // Nummer einer Textur

    // DrawingList
    boolean        MainList;                       // DrawingList verwalten
    GLint          MainListNr;                     // dito

    // Rest
    NodeType         MyNodeType;
    boolean        TesselationFlag, ModulateTextureFlag;
};

//=====================================================================
#endif // VRMLCONTROL_H
