// ===============================================
// File:                VRMLControl.cxx
// Description:         Parser for VRML97-files 
// Current:             Sept 28th 1999
// Author(s):           Thomas Kern
// ===============================================

/*
 *        thk VRML-Viewer (based on OpenGL)
 *  it shows static *.wrl files based on the VRML97 standard
 *
 *          Copyright (C) 1997-1999  Thomas Kern
 *                  All Rights Reserved
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * My email adress is: 
 *   Thomas.Kern@3Dsource.de 
 * 
 * You should find these files at: my homepage
 *   http://www.3Dsource.de
 *
 */

//=====================================================================
//	Description
//=====================================================================
// 
/*
	Here you can find the VRMLControl class.
  It is a wrapper class for the OpenGL objects in Objects.cxx.
  Some funftions are located in Tools.h/.cxx.

  You have to call the constructor with a *.wrl file name as a parameter.

  The constructor calls "ScanFile();", where
	- the file name and the directory names are separated (functions from Tools.h/.cxx)
	  + CheckFile(DateiName, "r")
	  + extractDir(char *complete, char *directory, char *file)
	- the directories are saved for file searching operations 
	  + scannedDir[dirCount]
	- are checks for the right file type (only unpacked VRML 2.0 are supported)
	  + CheckForType()
	- are the calls for parsing the file
	  + ScanForRootNodes();

  The method "ScanForRootNodes();"  
  - scans for "root nodes" (spezific VRML object entries) and generate node-names 
    + GenNodeName(NextNodeName);
  - scans for specific entries (not VRML, only my entries and comments)
    + #thk:NoTesselation, #thk:NoDecalTexture, #
  - scans for group or transformation entries:
    + if( !strcmp(nextString, "Group") )
        AddGroup(Ebene, Parent, NextNodeName);
    + else if( !strcmp(nextString, "Transform") )
        AddTransform(Ebene, Parent, NextNodeName);

  The method "AddGroup(Ebene, Parent, NextNodeName);" (objects from Objects.h/.cxx)
  - creates a GroupNode() object, name it and store the object in a list
    + ThisNode = new GroupNode();
    + ThisNode->SetName(ThisNodeName);
    + if(ObjectCopyCount < MAXNODES) {
        ObjectPtr[ObjectCount] = ThisNode;
  - specify the node as a root node (level 0) or a child (this node is encapsulated in another node entry)
    + RootNode[RootNodeCount] = ThisNode;
    + Parent->AddChildren( (VRMLNode*)ThisNode );
  - scans the node entry
    + ScanForBraces(Klammer2); // "{" increase a counter, "}" decrease the counter
    + adds a child node (keyword "children")
      AddChildren( Ebene, (GroupingNode*)ThisNode );
    + returns if "closing braces - open braces" = NULL

  The method "AddTransform(Ebene, Parent, NextNodeName);" (objects from Objects.h/.cxx)
  - creates a TransformNode() object
  - is looking for transform keywords ("center", "rotation", "translation", "scale" etc.)
  - calls private methods, i.e.
    + ThisNode->Rotation(Angle, Axle);

  The method "AddChildren(int Ebene, GroupingNode *Parent)"
  - counts the opening and closing brackets
    + ScanForBrackets(Klammer1); // []
  - is looking for keywords ("USE", "DEF", "Shape", "Group", Transform")
  - calls recursive methods
    + AddShape, AddGroup, AddTransform

  The method "AddShape(GroupingNode *Parent, char ThisNodeName[])" (objects from Objects.h/.cxx)
  - creates a ShapeNode() object
  - is looking for keywords and calls shape functions
    + KeyAppearance(ThisNode, NextNodeName);
    + KeyGeometry(ThisNode, NextNodeName);
  - or if there isn't an appearance entry, calls
    + AskForAppearance();
  
  The method "KeyAppearance(ShapeNode *MyShape, char ThisNodeName[])"
  - gets appearance entries from parent nodes ("USE")
  - scans for new appearance entries
    + AddAppearance(MyShape, ThisNodeName);

  The method "AddAppearance( ShapeNode *MyShape, char ThisNodeName[] )"
  - creates an AppearanceNode() object
  - adds an appearance entry to the parent shape node
    + MyShape->AddAppearance(ThisNode);
  - gets appearance entries from parent nodes ("USE")
    + ThisNode->AddMaterial, AddTexture, AddTextureTransform
  - scans the material, texture or texture transform entries
    + KeyMaterial, KeyTexture, KeyTextureTransform

  The method "AddMaterial(AppearanceNode *MyAppearance, char ThisNodeName[])"
  - creates a MaterialNode() object
  - adds a material entry to the parent appearance node
    + MyAppearance->AddMaterial(ThisNode);
  - scans and stores color and transparency values (OpenGL ready)

  The method "AddImageTexture(AppearanceNode *MyAppearance, char ThisNodeName[])"
  - creates an ImageTextureNode(TextureNr)
    + TextureNr = texture ID
    + AddTextureURL(url) (url = file name)
    + ScanTextureImage() (from Objects.h/.cxx)
    + loadPPMfile and loadJPEGfile (loading routines for PPM and JPEG from Tools.h/.cxx)
    + PrepareTexture() (OpenGL setup for this texture ID)

  The method "AddTextureTransform( AppearanceNode *MyAppearance, char ThisNodeName[] )"
  - creates a TextureTransformNode() for rotation, translation an scaling of a texture

  The method "KeyGeometry( ShapeNode *MyShape, char ThisNodeName[] )"
  - scans for different geometry keywords like
    + Cone, Cylinder, Sphere, Box (simple geometry objects)
    + IndexedFaceSet and IndexedLineSet (complex geometry objects)

  The Add-methods for Cone, Cylinder, Sphere, Box
  - create geometry node objects (from Objects.h/.cxx)
  - add a geometry entry to the parent appearance node
    + MyShape->AddGeometry( (GeometryNode*)ThisNode );

  The method AddLineSet
  - creates a Line Set object and adds a geometry entry to the parent appearance node
  - loads arrays for color, coordinates and the indices
    + KeyField(ThisNode, Color, NextNodeName);
    + KeyField(ThisNode, Coordinate, NextNodeName);
    + AddIndex(ThisNode, Color);
    + AddIndex(ThisNode, Coordinate);
  - calls PrepareForDraw() for setup 

  The method AddFaceSet
  - creating and array loading is like a Line Set Node
  - loads some extra settings (like TesselationFlag)
  - calls PrepareForDraw() for setup and calculating the normal vectors

*/


//=====================================================================
//      INCLUDES
//=====================================================================
//
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "Tools.H"
#include "VRMLControl.H"


// haeufig verwendete Variablen
static  const int   len3 = WORD_LENGHT;
static  char        dummy[LINEWIDE+1];
static  char        nextString[len3+1];          // \0-Zeichen muss auch noch rein

//=====================================================================
// Hilfsmethoden zur Verwaltung
//=====================================================================
//
// einen Namen generieren
void VRMLControl::GenNodeName(char newName[])
{
  sprintf( newName, "thk_InstanceName_%d", ObjectCount );
}

//---------------------------------------------------------------------
// Instanzen verwenden
int VRMLControl::UseNodeName()
{
  int    i, UseObjectNr = -1;
  char   name[len3+1];

  fscanf(file, "%s", nextString);                   // Namen lesen
  for(i = 0; i < ObjectCount; i++) {                // die Entsprechung suchen
    ObjectPtr[i]->GetName(name);
    if( !strcmp(nextString, name) ) {
      ObjectCopyPtr[ObjectCopyCount] = ObjectPtr[i];// zuordnen
      UseObjectNr = i;
    }
  }
  ObjectCopyCount++;
  return UseObjectNr;
}

//---------------------------------------------------------------------
// geschweifte Klammern zaehlen
void VRMLControl::ScanForBraces(int &Klammer2)
{
  if( !strcmp(nextString, "#") ||
      nextString[0] == '#' ) {                 // Kommentar
    fgets(dummy, LINEWIDE, file);
    fscanf(file, "%s", nextString);
  }
  else if( !strcmp(nextString, "{") ||
           nextString[0] == '{' )
    Klammer2++;                                // Bereich beginnt
  else if( !strcmp(nextString, "}") ||
           nextString[0] == '}' )
    Klammer2--;                                // Bereich endet
}

//---------------------------------------------------------------------
// eckige Klammern zaehlen
void VRMLControl::ScanForBrackets(int &Klammer1)
{
  if( !strcmp(nextString, "#")  ||
      nextString[0] == '#' ) {                 // Kommentar
    fgets(dummy, LINEWIDE, file);
    fscanf(file, "%s", nextString);
  }
  else if( !strcmp(nextString, "[") ||
           nextString[0] == '[' )
    Klammer1++;                                // Bereich beginnt
  else if( !strcmp(nextString, "]") ||
           nextString[0] == ']' )
    Klammer1--;                                // Bereich endet
}

//---------------------------------------------------------------------
// URL-Angabe auswerten
void VRMLControl::ScanURL(char DateiName[URL_SIZE])
{
  int       len1, len2, i, j;
  char      nextString[URL_SIZE];
  boolean   hadBracket = false;

  fscanf(file, "%s", nextString);          // cosmo designer ist nicht konform !
  if( !strcmp(nextString, "[") ) {
    hadBracket = true;
    fscanf(file, "%s", nextString);        //
  }

  len1 = strlen(nextString);
  len2 = URL_SIZE-1;
  i = j = 0;
  while( i < len1 ) {              // nicht ueber's Wortende hinaus
    if( (nextString[i] != '"') && (j < len2) ) {
      DateiName[j] = nextString[i];
      j++;
    }
    i++;
  }
  DateiName[j] = '\0';             // Wortende bestaetigen
  if(hadBracket) {
    fscanf(file, "%s", nextString);        // ]
  }

  // Kontrolle auf vernuenftige Angaben (nicht Internet-Bezug)
  if( DateiName[4] == ':' ) // file: oder http:
    DateiName[0] = '\0';
}


//=====================================================================
//  Methoden zum Erfassen der geometrischen Eigenschaften
//=====================================================================
//
// ein Wuerfel
void VRMLControl::AddBox( ShapeNode *MyShape, char ThisNodeName[] )
{
  int           i;
  boolean     stop = false;
  GLfloat     Geometry[3];
  BoxNode   *ThisNode = NULL;

  // zuerst ein neues Objekt anlegen
  ThisNode = new BoxNode();                      // neues Objekt
  if(ThisNode != NULL) {

    ThisNode->SetName(ThisNodeName);             // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    MyShape->AddGeometry( (GeometryNode*)ThisNode );
    
    // die Eintraege auswerten
    fscanf(file, "%s", nextString);             // oeffnende Klammer
    if( !strcmp(nextString, "{}") ) // nutze die Vorgabewerte
      stop = true;
    else {
      fscanf(file, "%s", nextString);
      if( !strcmp(nextString, "size") ) {
        i = 0;
        while( !stop && !feof(file) && (i < 3) ) {
          fscanf(file, "%s", nextString);
          if( !strcmp(nextString, "}") && nextString[0] != '}' )
            stop = true;
          else
            string2float(nextString, Geometry[i]); // die Zahlen in Float umwandeln
          i++;
        }
        if(i == 3)                               // drei Werte gefunden
          ThisNode->AddSize(Geometry);
      }
      while( nextString[0] != '}' && !feof(file) ) { // abschliessende Klammer lesen
        fscanf(file, "%s", nextString);
      }
    }
  }
}

//---------------------------------------------------------------------
// ein Kegel
void VRMLControl::AddCone( ShapeNode *MyShape, char ThisNodeName[] )
{
  boolean     stop = false;
  GLfloat     Geometry;
  ConeNode   *ThisNode = NULL; 
  
  // zuerst ein neues Objekt anlegen
  ThisNode = new ConeNode();                     // neues Objekt
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);             // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    MyShape->AddGeometry( (GeometryNode*)ThisNode );
    
    // die Eintraege auswerten
    fscanf(file, "%s", nextString);                     // oeffnende Klammer
    if( !strcmp(nextString, "{}") )     // nutze die Vorgabewerte
      stop = true;
    else {
      while( !stop && !feof(file) ) {
        fscanf(file, "%s", nextString);
        if( !strcmp(nextString, "}") ) {
          stop = true;
          break;
        }
        if( !strcmp(nextString, "bottomRadius") ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Geometry);
          ThisNode->AddRadius(Geometry);
        }
        else if( !strcmp(nextString, "height") ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Geometry);
          ThisNode->AddHeight(Geometry);
        }
        else if( !strcmp(nextString, "side") ) {
          fscanf(file, "%s", nextString);
          if( !strcmp(nextString, "FALSE") ) // TRUE ist Standard
          ThisNode->AddSide(false);
        }
        else if( !strcmp(nextString, "bottom") ) {
          fscanf(file, "%s", nextString);
          if( !strcmp(nextString, "FALSE") ) // TRUE ist Standard
            ThisNode->AddBottom(false);
        }
      }
    }
  }
}

//---------------------------------------------------------------------
// ein Zylinder
void VRMLControl::AddCylinder( ShapeNode *MyShape, char ThisNodeName[] )
{
  boolean     stop = false;
  GLfloat     Geometry;
  CylinderNode   *ThisNode = NULL;
  
  // zuerst ein neues Objekt anlegen
  ThisNode = new CylinderNode();                         // neues Objekt
  if(ThisNode != NULL ) {
    ThisNode->SetName(ThisNodeName);                     // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    MyShape->AddGeometry( (GeometryNode*)ThisNode );

    // die Eintraege auswerten
    fscanf(file, "%s", nextString);             // oeffnende Klammer
    // writeMessage(nextString); // thk-Debug
    if( !strcmp(nextString, "{}") ) // nutze die Vorgabewerte
      stop = true;
    else {
      while( !stop && !feof(file) ) {
        fscanf(file, "%s", nextString);
        if( !strcmp(nextString, "}") ) {
          stop = true;
          break;
        }
        
        if( !strcmp(nextString, "bottom") ) {
          fscanf(file, "%s", nextString);
          if( !strcmp(nextString, "FALSE") ) // TRUE ist Standard
            ThisNode->AddBottom(false);
        }
        else if( !strcmp(nextString, "height") ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Geometry);
          ThisNode->AddHeight(Geometry);
        }
        else if( !strcmp(nextString, "radius") ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Geometry);
          ThisNode->AddRadius(Geometry);
        }
        else if( !strcmp(nextString, "side") ) {
          fscanf(file, "%s", nextString);
          if( !strcmp(nextString, "FALSE") ) // TRUE ist Standard
            ThisNode->AddSide(false);
        }
        else if( !strcmp(nextString, "top") ) {
          fscanf(file, "%s", nextString);
          if( !strcmp(nextString, "FALSE") ) // TRUE ist Standard
            ThisNode->AddTop(false);
        }
      }
    }
  }
}

//---------------------------------------------------------------------
// eine Kugel
void VRMLControl::AddSphere( ShapeNode *MyShape, char ThisNodeName[] )
{
  GLfloat     Geometry;
  SphereNode   *ThisNode = NULL;
  
  // zuerst ein neues Objekt anlegen
  ThisNode = new SphereNode();                           // neues Objekt
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);                     // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    MyShape->AddGeometry( (GeometryNode*)ThisNode );

    // die Eintraege auswerten
    fscanf(file, "%s", nextString);           // oeffnende Klammer
    if( !!strcmp(nextString, "{}") && !feof(file) ) {
      fscanf(file, "%s", nextString);
      if( !strcmp(nextString, "radius") ) {
        fscanf(file, "%s", nextString);
        string2float(nextString, Geometry); // die Zahlen in Float umwandeln
        ThisNode->AddRadius(Geometry);
      }
      while( nextString[0] != '}' && !feof(file) ) { // abschliessende Klammer lesen
        fscanf(file, "%s", nextString);
      }
    }
  }
}


//=====================================================================
//  Polygonzuege, werden aus verschiedenen Nodes zusammengesetzt
//=====================================================================
//
// Indizes lesen
void VRMLControl::AddIndex( FaceSetNode *MyFaceSet,
                            NodeType     MyNodeType )
{
  int       i;
  boolean   stop = false;
  const int maxPoints = INDEX_MULT*MAXVERTICES; // Punkte werden mehrfach indiziert
  int       tempField[maxPoints];               //

  // Eintraege auswerten
  i = 0;
  while( !feof(file) && !stop && (i < maxPoints) ) {
    fscanf(file, "%s", nextString);              // naechstes Wort lesen
    if( !strcmp(nextString, "[") ||
        !strcmp(nextString, ",")   ) {
      fscanf(file, "%s", nextString);            // die Zahl nach dem Komma lesen
    }
    if( !strcmp(nextString, "]") )
      break;
    else if( !stop ) {
      sscanf(nextString, "%d", &tempField[i]);   // String zu int
      i++;                                       // das naechste Vertex
    }
  }

  if( i >= maxPoints-10 ) {
    sprintf(myMessageString,
            "Dieses Modell (%s) ist zu komplex.\nEs konnten nicht alle Punkte indiziert werden.",
            DateiName);
    writeMessage();
  }

  // jetzt Werte uebergeben (werden dort kopiert)
  if( MyNodeType == Color )
    MyFaceSet->AddColorIndex( tempField, i );
  else if( MyNodeType == Coordinate )
    MyFaceSet->AddCoordIndex( tempField, i );
  else if( MyNodeType == TextureCoordinate )
    MyFaceSet->AddTextureIndex( tempField, i );
}

//---------------------------------------------------------------------
// Koordinaten der Eckpunkte einlesen
void VRMLControl::AddField( FaceSetNode *MyFaceSet,
                            NodeType     MyNodeType,
                            char         ThisNodeName[] )
{
  int         i, Klammer1 = 0, Klammer2 = 0;
  boolean     stop = false;
  FieldNode  *ThisNode = NULL;
  const int   maxCoord = 3*MAXVERTICES; // Groesse des Datenfeldes
  GLfloat     tempField[maxCoord];      // Feld mit (2.000) Koordinaten (XYZ)

  // zuerst ein neues Objekt anlegen (wegen DEF/USE)
  ThisNode = new FieldNode();            // neues Objekt
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);     // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    
    while( !Klammer2 && !feof(file) ) {
      fscanf(file, "%s", nextString);
      // writeMessage(nextString); // thk-Debug
      ScanForBraces(Klammer2);                   // oeffnende Klammer {
    }
    
    // Syntax ueberpruefen
    fscanf(file, "%s", nextString);                // naechstes Wort lesen
    // writeMessage(nextString); // thk-Debug
    if( MyNodeType == Color && strcmp(nextString, "color") )
      stop = true;
    else if( MyNodeType == Coordinate && strcmp(nextString, "point") )
      stop = true;
    else if( MyNodeType == Normal && strcmp(nextString, "vector") )
      stop = true;
    else if( MyNodeType == TextureCoordinate && strcmp(nextString, "point") )
      stop = true;
    
    if( !stop ) {
      while( !Klammer1 && !feof(file) ) {
        fscanf(file, "%s", nextString);
        ScanForBrackets(Klammer1);                 // Klammer [
      }
    }
    
    // Eintraege auswerten
    i = 0;
    if( MyNodeType == TextureCoordinate ) {        // jeweils 2 Werte
      while( !feof(file) && !stop && (i < maxCoord-2) ) {
        fscanf(file, "%s", nextString);            // naechstes Wort lesen
        ScanForBrackets(Klammer1);                 // nur schliessende Klammer ]
        if( Klammer1 > 0 ) {
          if( !strcmp(nextString, ",") ) {
            fscanf(file, "%s", nextString);        // die Zahl nach dem Komma lesen
          }
          string2float(nextString, tempField[i  ]); // String zu Float
          fscanf(file, "%s", nextString);           // zweite Zahl lesen
          string2float(nextString, tempField[i+1]); // String zu Float
          i += 2;                                   // das naechste Vertex
        }
        else
          stop = true;
      }
    }
    else {                                       // jeweils 3 Werte
      while( !feof(file) && !stop && (i < maxCoord-3) ) {
        fscanf(file, "%s", nextString);            // naechstes Wort lesen
        ScanForBrackets(Klammer1);                 // nur schliessende Klammer ]
        if( Klammer1 > 0 ) {
          if( !strcmp(nextString, ",") ) {
            fscanf(file, "%s", nextString);        // die Zahl nach dem Komma lesen
          }
          string2float(nextString, tempField[i  ]); // String zu Float
          fscanf(file, "%s", nextString);           // zweite Zahl lesen
          string2float(nextString, tempField[i+1]); // String zu Float
          fscanf(file, "%s", nextString);           // dritte Zahl lesen
          string2float(nextString, tempField[i+2]); // String zu Float
          i += 3;                                   // das naechste Vertex
        }
        else
          stop = true;
      }
    }

    if( i >= maxCoord-20 ) {
      sprintf(myMessageString,
              "Dieses Modell (%s) ist zu komplex.\nEs konnten nicht alle Punkte erfasst werden.",
              DateiName);
      writeMessage();
    }
    
    while( Klammer2 && !feof(file) ) {
      fscanf(file, "%s", nextString);
      ScanForBraces(Klammer2);                   // schliessende Klammer }
    }
    
    // Feld und Typ uebergeben
    ThisNode->NewField(tempField, i);
    ThisNode->SetNodeTyp(MyNodeType);
    VFieldCount++;                               // jetzt kann hochgezaehlt werden
    
    if( MyNodeType == Color )
      MyFaceSet->AddColorField( ThisNode->ReturnFieldPtr() );
    else if( MyNodeType == Coordinate )
      MyFaceSet->AddCoordField( ThisNode->ReturnFieldPtr() );
    else if( MyNodeType == Normal )
      MyFaceSet->AddNormalField( ThisNode->ReturnFieldPtr() );
    else if( MyNodeType == TextureCoordinate )
      MyFaceSet->AddTextureField( ThisNode->ReturnFieldPtr() );
  }
}

//---------------------------------------------------------------------
// Verteiler auf Datenfeld
void VRMLControl::KeyField( FaceSetNode *MyFaceSet,
                            NodeType     MyNodeType,
                            char         ThisNodeName[] )
{
  int         UseObjectNr;
  FieldNode  *ThisNode = NULL;

  fscanf(file, "%s", nextString);              // naechstes Wort lesen
  if( !strcmp(nextString, "USE") ) {
    UseObjectNr = UseNodeName();
    if( UseObjectNr >= 0 ) {
      ThisNode = (FieldNode*)ObjectCopyPtr[UseObjectNr];
      if( ObjectCopyPtr[UseObjectNr]->GetNodeTyp() == Color )
        MyFaceSet->AddColorField( ThisNode->ReturnFieldPtr() );
      else if( ObjectCopyPtr[UseObjectNr]->GetNodeTyp() == Coordinate )
        MyFaceSet->AddCoordField( ThisNode->ReturnFieldPtr() );
      else if( ObjectCopyPtr[UseObjectNr]->GetNodeTyp() == Normal )
        MyFaceSet->AddNormalField( ThisNode->ReturnFieldPtr() );
      else if( ObjectCopyPtr[UseObjectNr]->GetNodeTyp() == TextureCoordinate )
        MyFaceSet->AddTextureField( ThisNode->ReturnFieldPtr() );
    }
  }
  else {
    if( !strcmp(nextString, "DEF") ) {
      fscanf(file, "%s", nextString);                            // den Namen lesen
      sprintf(ThisNodeName, "%s",nextString);
      fscanf(file, "%s", nextString);                            // Objektbezeichnung
    }

    // Syntax ueberpruefen
    if( MyNodeType == Color && !strcmp(nextString, "Color") )
      AddField(MyFaceSet, MyNodeType, ThisNodeName);
    else if( MyNodeType == Coordinate && !strcmp(nextString, "Coordinate") )
      AddField(MyFaceSet, MyNodeType, ThisNodeName);
    else if( MyNodeType == Normal && !strcmp(nextString, "Normal") )
      AddField(MyFaceSet, MyNodeType, ThisNodeName);
    else if( MyNodeType == TextureCoordinate && !strcmp(nextString, "TextureCoordinate") )
      AddField(MyFaceSet, MyNodeType, ThisNodeName);
  }
}

//---------------------------------------------------------------------
// hier erfolgt das Zusammensetzen des Polygonzugs
void VRMLControl::AddFaceSet( ShapeNode *MyShape, char ThisNodeName[] )
{
  int          Klammer2 = 0;
  boolean      stop = false;
  GLfloat      Angle;
  FaceSetNode *ThisNode = NULL;
  char         NextNodeName[len3+1];


  // zuerst ein neues Objekt anlegen
  ThisNode = new FaceSetNode();                          // neues Objekt
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);                     // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    MyShape->AddGeometry( (GeometryNode*)ThisNode );

    // Eintraege auswerten
    while( !feof(file) && !stop ) {
      fscanf(file, "%s", nextString);              // naechstes Wort lesen
      ScanForBraces(Klammer2);                           // Klammern
      GenNodeName(NextNodeName);                         // Namen generieren
      if( Klammer2 <= 0 ) {                         // {} bzw. DEF abfangen
        stop = true;
              break;
      }

      if( !strcmp(nextString, "color") )
        KeyField(ThisNode, Color, NextNodeName);
      else if( !strcmp(nextString, "coord") )
        KeyField(ThisNode, Coordinate, NextNodeName);
      else if( !strcmp(nextString, "normal") )
        KeyField(ThisNode, Normal, NextNodeName);
      else if( !strcmp(nextString, "texCoord") )
        KeyField(ThisNode, TextureCoordinate, NextNodeName);
      else if( !strcmp(nextString, "ccw") ) {
        fscanf(file, "%s", nextString);        
        if( !strcmp(nextString, "FALSE") ) // TRUE ist Standard
          ThisNode->SetCCW(false);
      }
      else if( !strcmp(nextString, "colorIndex") ) {
        AddIndex(ThisNode, Color);
      }
      else if( !strcmp(nextString, "colorPerVertex") ) {
        fscanf(file, "%s", nextString);
        if( !strcmp(nextString, "FALSE") ) // TRUE ist Standard
          ThisNode->SetColorPerVertex(false);
      }
      else if( !strcmp(nextString, "convex") ) {
        fscanf(file, "%s", nextString);
        if( !strcmp(nextString, "FALSE") ) { // TRUE ist Standard
          if( TesselationFlag )              // trotzdem Berechnung verhindern
            ThisNode->SetConvex(false);
        }
      }
      else if( !strcmp(nextString, "coordIndex") ) {
        AddIndex(ThisNode, Coordinate);
      }
      else if( !strcmp(nextString, "creaseAngle") ) {
        fscanf(file, "%s", nextString);
        string2float(nextString, Angle);
        ThisNode->SetCreaseAngle(Angle);
      }
      else if( !strcmp(nextString, "normalIndex") ) {
        AddIndex(ThisNode, Normal);
      }
      else if( !strcmp(nextString, "normalPerVertex") ) {
        fscanf(file, "%s", nextString);
        if( !strcmp(nextString, "FALSE") ) // TRUE ist Standard
          ThisNode->SetNormalPerVertex(false);
      }
      else if( !strcmp(nextString, "texCoordIndex") ) {
        AddIndex(ThisNode, TextureCoordinate);
      }
    }
    ThisNode->PrepareForDraw();                 // Vorgaben verarbeiten
  }
}

//---------------------------------------------------------------------
// hier erfolgt das Zusammensetzen des Linienzugs
void VRMLControl::AddLineSet( ShapeNode *MyShape, char ThisNodeName[] )
{
  int          Klammer2 = 0;
  boolean      stop = false;
  LineSetNode *ThisNode = NULL;
  char         NextNodeName[len3+1];


  // zuerst ein neues Objekt anlegen
  ThisNode = new LineSetNode();                                  // neues Objekt
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);                     // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    MyShape->AddGeometry( (GeometryNode*)ThisNode );

    // Eintraege auswerten
    while( !feof(file) && !stop ) {
      fscanf(file, "%s", nextString);              // naechstes Wort lesen
      // writeMessage(nextString); // thk-Debug
      ScanForBraces(Klammer2);                           // Klammern
      GenNodeName(NextNodeName);                                 // Namen generieren
      if( Klammer2 <= 0 ) {                        // {} bzw. DEF abfangen
        stop = true;
        break;
      }

      if( !strcmp(nextString, "color") )
        KeyField(ThisNode, Color, NextNodeName);
      else if( !strcmp(nextString, "coord") )
        KeyField(ThisNode, Coordinate, NextNodeName);
      else if( !strcmp(nextString, "colorIndex") )
        AddIndex(ThisNode, Color);
      else if( !strcmp(nextString, "colorPerVertex") ) {
        fscanf(file, "%s", nextString);
        if( !strcmp(nextString, "FALSE") ) // TRUE ist Standard
          ThisNode->SetColorPerVertex(false);
      }
      else if( !strcmp(nextString, "coordIndex") )
        AddIndex(ThisNode, Coordinate);
    }
    ThisNode->PrepareForDraw();                 // Vorgaben verarbeiten
  }
}


//=====================================================================
//  Methoden zur Beschreibung der Objekte
//=====================================================================
//
// Verteiler auf geometrische Objekte
void VRMLControl::KeyGeometry( ShapeNode *MyShape,
                               char ThisNodeName[] )
{
  int   UseObjectNr;

  fscanf(file, "%s", nextString);              // naechstes Wort lesen
  if( !strcmp(nextString, "USE") ) {
    UseObjectNr = UseNodeName();
    if( UseObjectNr >= 0 )                     // wirklich vorhanden
      MyShape->AddGeometry( (GeometryNode*)ObjectCopyPtr[UseObjectNr] );
  }
  else {
    if( !strcmp(nextString, "DEF") ) {
      fscanf(file, "%s", nextString);          // den Namen lesen
      sprintf(ThisNodeName, "%s",nextString);
      fscanf(file, "%s", nextString);          // Objektbezeichnung
    }

    if( !strcmp(nextString, "Box") )
      AddBox(MyShape, ThisNodeName);
    else if( !strcmp(nextString, "Cone") )
      AddCone(MyShape, ThisNodeName);
    else if( !strcmp(nextString, "Cylinder") )
      AddCylinder(MyShape, ThisNodeName);
    else if( !strcmp(nextString, "IndexedFaceSet") )
      AddFaceSet(MyShape, ThisNodeName);
    else if( !strcmp(nextString, "IndexedLineSet") )
      AddLineSet(MyShape, ThisNodeName);
    else if( !strcmp(nextString, "Sphere") )
      AddSphere(MyShape, ThisNodeName);
  }
}

//---------------------------------------------------------------------
// TextureTransform
void VRMLControl::AddTextureTransform( AppearanceNode *MyAppearance,
                                       char ThisNodeName[] )
{
  int           i, Klammer2 = 0;
  boolean       stop = false;
  GLfloat       Angle, Axle[2];
  TextureTransformNode  *ThisNode = NULL;

  // zuerst ein neues Objekt anlegen
  ThisNode = new TextureTransformNode();                 // neues Objekt
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);                     // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    MyAppearance->AddTextureTransform(ThisNode);

    // jetzt die Eintraege auswerten
    while( !feof(file) && !stop ) {
      fscanf(file, "%s", nextString);            // naechstes Wort lesen
      // writeMessage(nextString); // thk-Debug
      ScanForBraces(Klammer2);                   // Klammern
      if( Klammer2 <= 0 ) {                      // {} bzw. DEF abfangen
        stop = true;
        break;
      }

      if( !strcmp(nextString, "center") ) {
        for( i = 0; i < 2; i++ ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Axle[i]);
        }
        ThisNode->Center(Axle);
      }
      else if( !strcmp(nextString, "rotation") ) {
        fscanf(file, "%s", nextString);
        string2float(nextString, Angle);
        Angle *= 180/M_PI;                          // OpenGL verwendet Grad
        ThisNode->Rotation(Angle);
      }
      else if( !strcmp(nextString, "scale") ) {
        for( i = 0; i < 2; i++ ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Axle[i]);
        }
        ThisNode->Scale(Axle);
      }
      else if( !strcmp(nextString, "translation") ) {
        for( i = 0; i < 2; i++ ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Axle[i]);
        }
        ThisNode->Translation(Axle);
      }
    }
  }
}

//---------------------------------------------------------------------
// allgemeiner Verteiler Veraenderung der Texturkoordinaten
void VRMLControl::KeyTextureTransform( AppearanceNode *MyAppearance,
                                       char ThisNodeName[] )
{
  int  UseObjectNr;

  fscanf(file, "%s", nextString);              // naechstes Wort lesen
  if( !strcmp(nextString, "USE") ) {
    UseObjectNr = UseNodeName();
    if( UseObjectNr >= 0 )               // wirklich vorhanden
      MyAppearance->AddTextureTransform( (TextureTransformNode*)ObjectCopyPtr[UseObjectNr] );
  }
  else {
    if( !strcmp(nextString, "DEF") ) {
      fscanf(file, "%s", nextString);                    // den Namen lesen
      sprintf(ThisNodeName, "%s",nextString);
      fscanf(file, "%s", nextString);                    // Objektbezeichnung
    }
    if( !strcmp(nextString, "TextureTransform") )
      AddTextureTransform(MyAppearance, ThisNodeName);
  }
}

//---------------------------------------------------------------------
// ImageTexture aus Datei einbinden
void VRMLControl::AddImageTexture( AppearanceNode *MyAppearance,
                                   char ThisNodeName[] )
{
  int         Klammer2 = 0;
  boolean     stop = false;
  ImageTextureNode   *ThisNode = NULL;
  char        url[URL_SIZE];

  // zuerst ein neues Objekt anlegen
  ThisNode = new ImageTextureNode(TextureNr);            // neues Objekt
  if(ThisNode != NULL) {
    TextureNr++;
    ThisNode->SetName(ThisNodeName);                     // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    MyAppearance->AddTexture( (TextureNode*)ThisNode );          // anmelden

    // die Eintraege auswerten
    while( !feof(file) && !stop ) {
      fscanf(file, "%s", nextString);              // naechstes Wort lesen
      ScanForBraces(Klammer2);                     // Klammern
      if( Klammer2 <= 0 ) {                        // {} bzw. DEF abfangen
        stop = true;
        break;
      }

      if( !strcmp(nextString, "url") ) {
        ScanURL(url);
        ThisNode->AddTextureURL(url);
      }
      else if( !strcmp(nextString, "repeatS") ) {
        fscanf(file, "%s", nextString);
        if( !strcmp(nextString, "FALSE" ) )
          ThisNode->AddRepeatS(false); // TRUE ist Standard
      }
      else if( !strcmp(nextString, "repeatT") ) {
        fscanf(file, "%s", nextString);
        if( !strcmp(nextString, "FALSE" ) )
          ThisNode->AddRepeatT(false); // TRUE ist Standard
      }
    }    
    ThisNode->SetModulate(ModulateTextureFlag);
    ThisNode->PrepareTexture();                   // Textur vorbereiten
  }
}

//---------------------------------------------------------------------
// Textur einbinden (Verteiler)
void VRMLControl::KeyTexture( AppearanceNode *MyAppearance,
                              char ThisNodeName[] )
{
  int  UseObjectNr;

  fscanf(file, "%s", nextString);              // naechstes Wort lesen
  if( !strcmp(nextString, "USE") ) {
    UseObjectNr = UseNodeName();
    if( UseObjectNr >= 0 )               // wirklich vorhanden
      MyAppearance->AddTexture( (TextureNode*)ObjectCopyPtr[UseObjectNr] );
  }
  else {
    if( !strcmp(nextString, "DEF") ) {
      fscanf(file, "%s", nextString);                    // den Namen lesen
      sprintf(ThisNodeName, "%s",nextString);
      fscanf(file, "%s", nextString);                    // Objektbezeichnung
    }

    if( !strcmp(nextString, "ImageTexture") )
      AddImageTexture(MyAppearance, ThisNodeName);
  }
}

//---------------------------------------------------------------------
// Farbe und Relexionseigenschaften
void VRMLControl::AddMaterial( AppearanceNode *MyAppearance,
                               char ThisNodeName[] )
{
  int           i, Klammer2 = 0;
  boolean     stop = false;
  GLfloat     Intensity;
  GLclampf    Color3[3];
  MaterialNode  *ThisNode = NULL;

  // zuerst ein neues Objekt anlegen
  ThisNode = new MaterialNode();         // neues Objekt
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);     // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    MyAppearance->AddMaterial(ThisNode);         // anmelden

    // die Eintraege auswerten
    while( !feof(file) && !stop ) {
      fscanf(file, "%s", nextString);                    // naechstes Wort lesen
      ScanForBraces(Klammer2);                                 // Klammern
      if( Klammer2 <= 0 ) {                         // {} bzw. DEF abfangen
        stop = true;
        break;
      }

      if( !strcmp(nextString, "ambientIntensity") ) {
        fscanf(file, "%s", nextString);
        string2float(nextString, Intensity);
        ThisNode->AddAmbient( (GLclampf)(Intensity) );
      }
      else if( !strcmp(nextString, "diffuseColor") ) {
        for(i = 0; i < 3; i++) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Intensity);
          Color3[i] = (GLclampf)(Intensity);
        }
        if( i == 3 )
          ThisNode->AddDiffuse(Color3);
      }
      else if( !strcmp(nextString, "emissiveColor") ) {
        //AddEmissive(    GLclampf Color[3] );
      }
      else if( !strcmp(nextString, "shininess") ) {
        //AddShininess(   GLclampf Intensity);
      }
      else if( !strcmp(nextString, "specularColor") ) {
        //AddSpecular(       GLclampf Color[3] );
      }
      else if( !strcmp(nextString, "transparency") ) {
        fscanf(file, "%s", nextString);
        string2float(nextString, Intensity);
        ThisNode->AddTransparency( (GLclampf)(Intensity) );
      }
    }
  }
}

//---------------------------------------------------------------------
// allgemeiner Verteiler auf Farbe und Relexionseigenschaften
void VRMLControl::KeyMaterial( AppearanceNode *MyAppearance,
                               char ThisNodeName[] )
{
  int  UseObjectNr;
  
  fscanf(file, "%s", nextString);              // naechstes Wort lesen
  if( !strcmp(nextString, "USE") ) {
    UseObjectNr = UseNodeName();
    if( UseObjectNr >= 0 )               // wirklich vorhanden
      MyAppearance->AddMaterial( (MaterialNode*)ObjectCopyPtr[UseObjectNr] );
  }
  else {
    if( !strcmp(nextString, "DEF") ) {
      fscanf(file, "%s", nextString);                    // den Namen lesen
      sprintf(ThisNodeName, "%s",nextString);
      fscanf(file, "%s", nextString);                    // Objektbezeichnung
    }
    if( !strcmp(nextString, "Material") )
      AddMaterial(MyAppearance, ThisNodeName);
  }
}

//---------------------------------------------------------------------
// wiederverwendbare Node mit allen Materialeigenschaften
void VRMLControl::AddAppearance( ShapeNode *MyShape,
                                 char ThisNodeName[] )
{
  int            UseObjectNr, Klammer2 = 0;
  boolean        stop = false;
  AppearanceNode *ThisNode = NULL;
  NodeType       MyNodeTyp;
  char           NextNodeName[len3+1];

  // zuerst ein neues Objekt anlegen
  ThisNode = new AppearanceNode();                               // neues Objekt
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);                             // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    MyShape->AddAppearance(ThisNode);

    // Eintraege auswerten
    while( !feof(file) && !stop ) {
      fscanf(file, "%s", nextString);                    // naechstes Wort lesen
      ScanForBraces(Klammer2);                           // Klammern
      GenNodeName(NextNodeName);                         // Namen generieren
      if( Klammer2 <= 0 ) {                              // {} bzw. DEF abfangen
        stop = true;
        break;
      }

      if( !strcmp(nextString, "DEF") ) {
        fscanf(file, "%s", nextString);                            // den Namen lesen
        sprintf(NextNodeName, "%s",nextString);
      }
      else if( !strcmp(nextString, "USE") ) {
        UseObjectNr = UseNodeName();
        if( UseObjectNr >= 0 ) {                                   // wirklich vorhanden
          MyNodeTyp = (NodeType)ObjectCopyPtr[UseObjectNr]->GetNodeTyp();
          if( MyNodeTyp == Material )
            ThisNode->AddMaterial( (MaterialNode*)ObjectCopyPtr[UseObjectNr] );
          else if( (MyNodeTyp == ImageTexture) || (MyNodeTyp == PixelTexture) )
            ThisNode->AddTexture( (TextureNode*)ObjectCopyPtr[UseObjectNr] );
          else if( MyNodeTyp == TextureTransform )
            ThisNode->AddTextureTransform( (TextureTransformNode*)ObjectCopyPtr[UseObjectNr] );
        }
      }
      else if( !strcmp(nextString, "material") )
        KeyMaterial(ThisNode, NextNodeName);
      else if( !strcmp(nextString, "texture") )
        KeyTexture(ThisNode, NextNodeName);
      else if( !strcmp(nextString, "textureTransform") )
        KeyTextureTransform(ThisNode, NextNodeName);
    }
  }
}

//---------------------------------------------------------------------
// allgemeiner Verteiler auf Materialeigenschaften
void VRMLControl::KeyAppearance( ShapeNode *MyShape,
                                 char ThisNodeName[] )
{
  int  UseObjectNr;

  fscanf(file, "%s", nextString);              // naechstes Wort lesen
  if( !strcmp(nextString, "USE") ) {
    UseObjectNr = UseNodeName();
    if( UseObjectNr >= 0 )
      MyShape->AddAppearance( (AppearanceNode*)ObjectCopyPtr[UseObjectNr] );
  }
  else {
    if( !strcmp(nextString, "DEF") ) {
      fscanf(file, "%s", nextString);                    // den Namen lesen
      sprintf(ThisNodeName, "%s",nextString);
      fscanf(file, "%s", nextString);                    // Objektbezeichnung
    }
    if( !strcmp(nextString, "Appearance") )
      AddAppearance(MyShape, ThisNodeName);
  }
}


//=====================================================================
//  Methoden zum Erzeugen und Verwalten der Objekte
//=====================================================================
//
// Koerperbeschreibung lesen (enthaelt Geoemtrie, Farbe und Texturen)
void VRMLControl::AddShape( GroupingNode *Parent,
                            char ThisNodeName[] )
{
  int         Klammer2 = 0;
  boolean     stop = false;
  ShapeNode  *ThisNode = NULL;
  char        NextNodeName[len3+1];
  
  // zuerst ein neues Objekt anlegen
  ThisNode = new ShapeNode();
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    Parent->AddChildren( (VRMLNode*)ThisNode );    // beim Aufrufer anmelden

    // die Eintraege auswerten
    while( !feof(file) && !stop ) {
      fscanf(file, "%s", nextString);                    // naechstes Wort lesen
      ScanForBraces(Klammer2);                     // Klammern
      GenNodeName(NextNodeName);                         // Namen generieren
      if( Klammer2 <= 0 ) {                         // {}
        stop = true;
        break;
      }

      if( !strcmp(nextString, "appearance") )
        KeyAppearance(ThisNode, NextNodeName);
      else if( !strcmp(nextString, "geometry") )
        KeyGeometry(ThisNode, NextNodeName);
    }
    
    // Shape muss nicht zwingend eine ApperanceNode haben
    ThisNode->AskForAppearance();
  }
}

//---------------------------------------------------------------------
// Children-Eintrag auswerten (Verteiler)
void VRMLControl::AddChildren( int Ebene, GroupingNode *Parent )
{
  int        UseObjectNr, Klammer1 = 0;
  boolean    stop = false;
  char      NextNodeName[len3+1];

  // die Eintraege auswerten
  while( !feof(file) && !stop ) {
    fscanf(file, "%s", nextString);                    // naechstes Wort lesen
    ScanForBrackets(Klammer1);
    GenNodeName(NextNodeName);
    if( Klammer1 <= 0 )                              // [] bzw. DEF abfangen
      stop = true;

    if( !strcmp(nextString, "USE") ) {
      UseObjectNr = UseNodeName();
      if( UseObjectNr >= 0 )                     // wirklich vorhanden
        Parent->AddChildren( ObjectCopyPtr[UseObjectNr] );
    }
    else {
      if( !strcmp(nextString, "DEF") ) {
        fscanf(file, "%s", nextString);                    // den Namen lesen
        sprintf(NextNodeName, "%s",nextString);
        fscanf(file, "%s", nextString);                    // Objektbezeichnung
      }

      if( !strcmp(nextString, "Shape") )
        AddShape(Parent, NextNodeName);
      else if( !strcmp(nextString, "Group") )
        AddGroup(Ebene+1, Parent, NextNodeName);
      else if( !strcmp(nextString, "Transform") )
        AddTransform(Ebene+1, Parent, NextNodeName);
      else if( !strcmp(nextString, "WorldInfo") ) {
        while( !!strcmp(nextString, "}") ) {
          fscanf(file, "%s", nextString);
        }
      }
    }
  }
}

//---------------------------------------------------------------------
// Transform-Node auswerten
void VRMLControl::AddTransform( int Ebene,
                                GroupingNode *Parent,
                                char ThisNodeName[] )
{
  int           i, Klammer2 = 0;
  boolean       stop = false;
  GLfloat       Angle, Axle[3];
  TransformNode  *ThisNode = NULL;

  // zuerst ein neues Objekt anlegen
  ThisNode = new TransformNode();                        // neues Objekt
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);                     // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }

    // den Speziallfall der Ebene Null beruecksichtigen
    if( !Ebene ) {
      RootNode[RootNodeCount] = ThisNode;                // steuerndes Objekt
      RootNodeCount++;
    }
    else
      Parent->AddChildren( (VRMLNode*)ThisNode );    // beim Aufrufer anmelden

    // jetzt die Eintraege auswerten
    while( !feof(file) && !stop ) {
      fscanf(file, "%s", nextString);                   // naechstes Wort lesen
      ScanForBraces(Klammer2);                          // Klammern
      if( Klammer2 <= 0 ) {                             // {} bzw. DEF abfangen
        stop = true;
        break;
      }

      if( !strcmp(nextString, "center") ) {
        for( i = 0; i < 3; i++ ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Axle[i]);
        }
        ThisNode->Center(Axle);
      }
      else if( !strcmp(nextString, "rotation") ) {
        for( i = 0; i < 3; i++ ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Axle[i]);
        }
        fscanf(file, "%s", nextString);
        string2float(nextString, Angle);
        Angle *= 180/M_PI;                          // OpenGL verwendet Grad
        ThisNode->Rotation(Angle, Axle);
      }
      else if( !strcmp(nextString, "rotationD") ) {
        for( i = 0; i < 3; i++ ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Axle[i]);
        }
        fscanf(file, "%s", nextString);
        string2float(nextString, Angle);
        ThisNode->Rotation(Angle, Axle);
      }
      else if( !strcmp(nextString, "scale") ) {
        for( i = 0; i < 3; i++ ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Axle[i]);
        }
        ThisNode->Scale(Axle);
      }
      else if( !strcmp(nextString, "scaleOrientation") ) {
        for( i = 0; i < 3; i++ ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Axle[i]);
        }
        fscanf(file, "%s", nextString);
        string2float(nextString, Angle);
        Angle *= 180/M_PI;                          // OpenGL verwendet Grad
        ThisNode->ScaleOrientation(Angle, Axle);
      }
      else if( !strcmp(nextString, "translation") ) {
        for( i = 0; i < 3; i++ ) {
          fscanf(file, "%s", nextString);
          string2float(nextString, Axle[i]);
        }
        ThisNode->Translation(Axle);
      }
      else if( !strcmp(nextString, "children") )
        AddChildren( Ebene, (GroupingNode*)ThisNode );
    }
  }
}

//---------------------------------------------------------------------
// Group-Node auswerten
void VRMLControl::AddGroup( int Ebene,
                            GroupingNode *Parent,
                            char ThisNodeName[] )
{
  int         Klammer2 = 0;
  boolean     stop = false;
  GroupNode  *ThisNode = NULL;

  // zuerst ein neues Objekt anlegen
  ThisNode = new GroupNode();                            // neues Objekt
  if(ThisNode != NULL) {
    ThisNode->SetName(ThisNodeName);                     // Namen zuweisen
    if(ObjectCopyCount < MAXNODES) {
      ObjectPtr[ObjectCount] = ObjectCopyPtr[ObjectCopyCount] = ThisNode;
      ObjectCount++; ObjectCopyCount++;
    }
    
    // den Speziallfall der Ebene Null beruecksichtigen
    if( !Ebene ) {
      RootNode[RootNodeCount] = ThisNode;                // steuerndes Objekt
      RootNodeCount++;
    }
    else
      Parent->AddChildren( (VRMLNode*)ThisNode );    // beim Aufrufer anmelden
    
    // jetzt die Eintraege auswerten
    while( !feof(file) && !stop ) {
      fscanf(file, "%s", nextString);            // naechstes Wort lesen
      ScanForBraces(Klammer2);                   // "{" increase a counter, "}" decrease the counter

      if( !strcmp(nextString, "children") )
        AddChildren( Ebene, (GroupingNode*)ThisNode );

      if( Klammer2 <= 0 )                        // {}, counter = 0 or more closing braces than open braces
        stop = true;
    }
  }
}

//---------------------------------------------------------------------
// die einzelnen Eintraege auswerten (auch als Verteiler)
void VRMLControl::ScanForRootNodes()
{
  int           Ebene;
  GroupingNode *Parent = NULL;
  char          NextNodeName[len3+1];

  // jetzt die Eintraege auswerten
  while( !feof(file) ) {
    Ebene = 0;
    GenNodeName(NextNodeName);                         // eindeutiger Name
    fscanf(file, "%s", nextString);                    // naechstes Wort lesen

    if( !strcmp(nextString, "#thk:NoTesselation") ) {
      showMessage("NoTesselation flag found");
      TesselationFlag = false;                         // eigenes Flag
      fgets(dummy, LINEWIDE, file);
      continue;
    }
    if( !strcmp(nextString, "#thk:NoDecalTexture") ) {
      showMessage("NoDecalTexture flag found");
      ModulateTextureFlag = true;
      fgets(dummy, LINEWIDE, file);
      continue;
    }      
    if( !strcmp(nextString, "#") ||
        nextString[0] == '#' ) {                  // Kommentar
      fgets(dummy, LINEWIDE, file);
      fscanf(file, "%s", nextString);
    }
    if( !strcmp(nextString, "Group") )
      AddGroup(Ebene, Parent, NextNodeName);
    else if( !strcmp(nextString, "Transform") )
      AddTransform(Ebene, Parent, NextNodeName);
  }
}

//---------------------------------------------------------------------
// File-Typ erkennen
int VRMLControl::CheckForType()
{
  int      i, found = 0, Typ;
  char    *vergleich1[] = {"#VRML", "VRML", "V2.0", "utf8"};
  int      eintraege1 = 4; // Anzahl der Woerter in vergleich
  
  const int len3 = WORD_LENGHT;
  char  nextString[len3+1];           // \0-Zeichen muss auch noch rein
  for(i = 0; i < len3; i++)
    nextString[i] = '0';              // vordefinierter Wert

  Typ = 0;
  while( !found && !feof(file) ) { // ersten Treffer suchen
    fscanf(file, "%s", nextString);
    if( !strcmp(nextString, vergleich1[0]) ||
        !strcmp(nextString, vergleich1[1])    )
      found++;                  // Kommentarzeichen gilt ebenfalls
  }
  if(found) {
    for(i = 2; i < eintraege1; i++) {
      fscanf(file, "%s", nextString);
      if( !strcmp(nextString, vergleich1[i]) )
        found++;
    }
    if( found == 3 )
      Typ = 2;                   // VRML 2
    else {
      // Scan for Inventor etc.
    }
    fgets(dummy, LINEWIDE, file); // Rest der Zeile ignorieren
  }
  return Typ;
}

//---------------------------------------------------------------------
// gesamte Datei einlesen (Verteilerfunktion)
void VRMLControl::ScanFile()
{
  file  = NULL;

  if( CheckFile(DateiName, "r") )
    file = fopen(DateiName, "r");

  if(file == NULL) {
    sprintf(myMessageString, "Objektbeschreibung '%s' nicht gefunden.", DateiName);
    writeMessage();
  }
  else {
    fgets(nextString, 6, file);
    if( nextString[2] == '\b')
      showMessage("Datei ist binaer. Abbruch.");
    else {
      fclose(file);
      file = fopen(DateiName, "r");
      switch( CheckForType() ) {
      case 0:
              sprintf(myMessageString, "unbekanntes Dateiformat: %s", DateiName);
              writeMessage();
              break;
      case 2:
              sprintf(myMessageString, "Lade VRML 2.0 - Datei: '%s'", DateiName);
              writeMessage();
              while( !feof(file) )
                ScanForRootNodes();
              break;
      }
    }
    fclose(file);
  }  
}


//=====================================================================
//  folgende Methoden lassen sich von aussen aufrufen
//=====================================================================
//
// Texturen zu- oder abschalten
void VRMLControl::SetTexture(boolean On)
{
  for(int i = 0; i < ObjectCount; i++)
    ObjectPtr[i]->SetTexture(On);
}

//---------------------------------------------------------------------
// Zeichnung ausfuehren
void VRMLControl::DrawObject()
{
  for(int i = 0; i < RootNodeCount; i++) {
    glPushMatrix();
      RootNode[i]->DrawNode();
    glPopMatrix();
  }
}

//---------------------------------------------------------------------
// als (einmalig berechnete) Liste zeichen
void VRMLControl::DrawAsList()
{
  if( !MainList ) {
    MainListNr = glGenLists(1); // neue Listenummer erzeugen
    glNewList(MainListNr, GL_COMPILE_AND_EXECUTE); // enthaelt alle Teilmodelle
    DrawObject();
    glEndList();
    MainList = true;
  }
  else
    glCallList(MainListNr);
}

//---------------------------------------------------------------------
// VRMLControl: Destruktor
VRMLControl::~VRMLControl()
{
  int i;

  if(MainList)
    glDeleteLists(MainListNr, 1);
  
  for(i = 0; i < ObjectCount; i++)
    delete ObjectPtr[i];
}

//---------------------------------------------------------------------
// VRMLControl: Konstruktor
VRMLControl::VRMLControl(char Datei[])
{
  file = NULL;
  DateiName[0] = '\0';

  sprintf(DateiName, "%s", Datei);      // VRML-Datei
  ObjectCount = ObjectCopyCount = 0;    // keine VRML-Objekte
  RootNodeCount   = 0;                  // Darstellungsfaehige Objekte
  VFieldCount     = 0;                  // keine Vertex-Arrays
  TextureNr       = 1;                  // Texturname Null ist Standard
  MainList        = false;              // keine Display-List
  MainListNr      = 0;
  TesselationFlag  = true;              // Tesselation nutzen ??
  ModulateTextureFlag = false;          // Farbe unter Textur beruecksichtigen ?

  ScanFile();                   // erzeugt ein Modell aus Datei "DateiName"
}

